#!/usr/bin/env julia

###
#COPYRIGHT (c) 2018 THE OHIO STATE UNIVERSITY
#ALL RIGHTS RESERVED
#
#PERMISSION IS GRANTED TO USE, COPY, CREATE DERIVATIVE WORKS AND
#REDISTRIBUTE THIS SOFTWARE AND SUCH DERIVATIVE WORKS FOR NONCOMMERCIAL
#EDUCATION AND RESEARCH PURPOSES, SO LONG AS NO FEE IS CHARGED, AND SO
#LONG AS THE COPYRIGHT NOTICE ABOVE, THIS GRANT OF PERMISSION, AND THE
#DISCLAIMER BELOW APPEAR IN ALL COPIES MADE; AND SO LONG AS THE NAME OF
#THE OHIO STATE UNIVERSITY IS NOT USED IN ANY ADVERTISING OR PUBLICITY
#PERTAINING TO THE USE OR DISTRIBUTION OF THIS SOFTWARE WITHOUT
#SPECIFIC, WRITTEN PRIOR AUTHORIZATION.
#
#THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE OHIO
#STATE UNIVERSITY AS TO ITS FITNESS FOR ANY PURPOSE, AND WITHOUT
#WARRANTY BY THE OHIO STATE UNIVERSITY OF ANY KIND, EITHER EXPRESS OR
#IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF
#MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE OHIO STATE
#UNIVERSITY SHALL NOT BE LIABLE FOR ANY DAMAGES, INCLUDING SPECIAL,
#INDIRECT, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY
#CLAIM ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE,
#EVEN IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
#DAMAGES.
###

function check_constrain(x)
	thr=0.01
	if length(x) !=6
		@printf(STDERR,"check_constrain(): must be 6 elements\n")
		exit(2)
	end
	if x[1] >thr && x[2]<(-thr) && x[3]>thr && x[4]>thr && x[5] >thr && x[6]>thr
		return "YES"
	else
		return "NO"
	end
end

function score_func(x,psml)

	if length(x) !=6
		@printf(STDERR,"score_func(): length: must be 6 elements\n")
		exit(4)
	end

	if length(psml) <1
		@printf(STDERR,"score_func(): psml is empty\n")
		exit(4)
	end

	(a,b,sAt,sTt,sGt,sCt)=x
	result=0
	seq_score=0
	seq=""
	count=NaN

	r=NaN
	r_offset=1e-10

	for j in 1:length(psml)
		seq_score=0
		(seq,count) = split(psml[j])
		nts = split(seq,"")
		count = int(count)
		for i in 1:length(nts)-1
			r = 0.5*(erf(a*(i-1+b))+1) + r_offset
			if nts[i] == "A"
				seq_score += -r*sAt
			elseif nts[i] == "T"
				seq_score += -r*sTt
			elseif nts[i] == "G"
				seq_score += -r*sGt
			elseif nts[i] == "C"
				seq_score += -r*sCt
			else
				@printf(STDERR,"score_func(): unknown char %s in seq %s\n",nts[i],seq)
				exit(5)
			end
		end

		r = 0.5*( erf(a*(length(nts)-1+b))+1 ) + r_offset

		if nts[end] == "A"
			seq_score += log(1-exp(-r*sAt))
		elseif nts[end] == "T"
			seq_score += log(1-exp(-r*sTt))
		elseif nts[end] == "G"
			seq_score += log(1-exp(-r*sGt))
		elseif nts[end] == "C"
			seq_score += log(1-exp(-r*sCt))
		else
			@printf(STDERR,"score_func(): unknown last char %s\n",nts[end])
			exit(6)
		end

		result += seq_score*count
	end

	return result
end

function energy_of(x,psml)
	return -score_func(x,psml)
end

function gen_new_state_int(x)
	old_x= x
	new_x=fill(NaN,6)

	a_step=0.1
	a_step_small=0.01
	s_step=0.1
	s_step_small=0.01

	if old_x[1] > a_step*2
		new_x[1] = old_x[1] + (int(9*rand())*a_step-4*a_step)
	else
		new_x[1] = old_x[1] + (int(9*rand())*a_step_small-4*a_step_small)
	end

	new_x[2] = old_x[2]

	if old_x[3] > s_step*2
		new_x[3] = old_x[3] + (int(9*rand())*s_step-4*s_step)
	else
		new_x[3] = old_x[3] + (int(9*rand())*s_step_small-4*s_step_small)
	end

	if old_x[4] > s_step*2
		new_x[4] = old_x[4] + (int(9*rand())*s_step-4*s_step)
	else
		new_x[4] = old_x[4] + (int(9*rand())*s_step_small-4*s_step_small)
	end

	if old_x[5] > s_step*2
		new_x[5] = old_x[5] + (int(9*rand())*s_step-4*s_step)
	else
		new_x[5] = old_x[5] + (int(9*rand())*s_step_small-4*s_step_small)
	end

	if old_x[6] > s_step*2
		new_x[6] = old_x[6] + (int(9*rand())*s_step-4*s_step)
	else
		new_x[6] = old_x[6] + (int(9*rand())*s_step_small-4*s_step_small)
	end

	return new_x
end

#***********************************************************************
#                            MAIN PROGRAM
#***********************************************************************

if length(ARGS) != 1
	@printf(STDERR,"Usage: RiboProP_fit pool\n")
	@printf(STDERR,"       pool: plain text file of the 3' pool\n")
	exit(1)
end

psmf = ARGS[1]

psml=[]
for line in eachline(open(psmf))
	if ismatch(r"^[^#]",line)
		psml=[psmLines,chomp(line)]
	end
end

# initial model parameters
a   =
b   =
sAt =
sTt =
sGt =
sCt =

# simulated annealing parameters

# initial temperature, e.g. 3000
t_init = 
# minimum temperature, e.g. 1
t_min  =
# cooling rate, e.g. 0.99
mu_t   = 
# iterations per temperature, e.g. 300
niter_max = 

# simulated annealing routines
t=NaN
niter=NaN
energy=NaN
new_energy=NaN

x_init=[a,b,sAt,sTt,sGt,sCt]

if check_constrain(x_init) == "NO"
	@printf(STDERR,"main(): initialization failed\n")
	exit(3)
end

x=x_init
t=t_init

while((t*=mu_t)>t_min)
	niter=0

	while niter<niter_max

		new_x = gen_new_state_int(x)

		if check_constrain(new_x) == "NO"
			continue
		end

		energy = energy_of(x,psml)
		new_energy = energy_of(new_x,psml)

		if new_energy < energy
			x = new_x
		elseif rand() < exp( (energy - new_energy)/t)
			x = new_x
		end

		niter +=1
	end
end
