#!/usr/bin/env julia

###
#COPYRIGHT (c) 2018 THE OHIO STATE UNIVERSITY
#ALL RIGHTS RESERVED
#
#PERMISSION IS GRANTED TO USE, COPY, CREATE DERIVATIVE WORKS AND
#REDISTRIBUTE THIS SOFTWARE AND SUCH DERIVATIVE WORKS FOR NONCOMMERCIAL
#EDUCATION AND RESEARCH PURPOSES, SO LONG AS NO FEE IS CHARGED, AND SO
#LONG AS THE COPYRIGHT NOTICE ABOVE, THIS GRANT OF PERMISSION, AND THE
#DISCLAIMER BELOW APPEAR IN ALL COPIES MADE; AND SO LONG AS THE NAME OF
#THE OHIO STATE UNIVERSITY IS NOT USED IN ANY ADVERTISING OR PUBLICITY
#PERTAINING TO THE USE OR DISTRIBUTION OF THIS SOFTWARE WITHOUT
#SPECIFIC, WRITTEN PRIOR AUTHORIZATION.
#
#THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE OHIO
#STATE UNIVERSITY AS TO ITS FITNESS FOR ANY PURPOSE, AND WITHOUT
#WARRANTY BY THE OHIO STATE UNIVERSITY OF ANY KIND, EITHER EXPRESS OR
#IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF
#MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE OHIO STATE
#UNIVERSITY SHALL NOT BE LIABLE FOR ANY DAMAGES, INCLUDING SPECIAL,
#INDIRECT, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY
#CLAIM ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE,
#EVEN IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
#DAMAGES.
###

function cal_cov(a3_cov,seq,para,genelen,ext,mat)

	tp_cov = zeros(1,length(a3_cov))

	if(ext<=0)
		@printf(STDERR,"cal_cov(): extended_len is less than zero\n")
		exit(2)
	end

	for i = 1:genelen
		for j= 0:ext
			tp_cov[i+j] += a3_cov[i]*mat[i,j+1]
		end
	end

	return tp_cov
end

function prob_seq(nts,para)

	(a,b,sAt,sTt,sGt,sCt) = para

	r_offset=1e-10
	seq_score = 1

	for i in 1:length(nts)-1
		r = erf(a*(i-1+b))+1+r_offset

		if nts[i] == 'A'
			seq_score *= exp(-r*sAt)
		elseif nts[i] == 'T'
			seq_score *= exp(-r*sTt)
		elseif nts[i] == 'G'
			seq_score *= exp(-r*sGt)
		elseif nts[i] == 'C'
			seq_score *= exp(-r*sCt)
		else
			@printf(STDERR,"prob_seq(): unknown char (",nts[i],") in seq (",seq,")\n")
			exit(5)
		end
	end

	r = erf(a*(length(nts)-1+b))+1+r_offset

	if nts[end] == 'A'
		seq_score *= (1-exp(-r*sAt))
	elseif nts[end] == 'T'
		seq_score *= (1-exp(-r*sTt))
	elseif nts[end] == 'G'
		seq_score *= (1-exp(-r*sGt))
	elseif nts[end] == 'C'
		seq_score *= (1-exp(-r*sCt))
	else
		@printf(STDERR,"prob_seq(): unknown last char %s\n",nts[end])
		exit(6)
	end

	return seq_score
end

#**********************************************************************
#                           MAIN PROGRAM
#**********************************************************************

#---------------------  set up program arguments  ---------------------

# The real 3' coverage, e.g. [12 23 34 53 12 3 4 ...]
tp_cov_real = 
tp_cov_real = tp_cov_real/mean(tp_cov_real)

# the corresponding nucleotide sequence, e.g. "ATGCGTCCTAGGACC..."
seq = 

# model parameters in the order of [a b sAt sTt sGt sCt]
para = 

# length of the gene, integer
genelen = 

#----------------------------------------------------------------------
extended_len = length(tp_cov_real)-genelen

seq_score_mat=zeros(genelen,extended_len+1)

for i = 1:genelen
	for j= 1:extended_len+1
		seq_score_mat[i,j] = prob_seq(seq[i:i+j],para)
	end
end
#----------------------------------------------------------------------

if(length(seq)!= length(tp_cov_real)+1)
	@printf(STDERR,"length of sequence is not length of coverages + 1\n")
	exit(2)
end

dimen=length(tp_cov_real)


# simulated annealing parameters

# initial temperature, e.g. 3000
t_init = 
# minimum temperature, e.g. 1
t_min  =
# cooling rate, e.g. 0.99
mu_t   = 
# iterations per temperature, e.g. 300
niter_max = 

# other parameters to fine-tune the algorithm
prefac=100
sa_prefac=1e5
multistep=2


# simulated annealing routines

a3_cov_init = rand(1,dimen)
a3_cov = a3_cov_init

tp_cov = []

t=t_init

while (t*= mu_t) > t_min
	niter=0

	while niter <niter_max

		a3_cov = a3_cov/mean(a3_cov)*prefac
		new_a3_cov = a3_cov + (2*rand(1,dimen)-1)*multistep

		for i in find(x->x<0,new_a3_cov)
			new_a3_cov[i]=0
		end

		a3_cov = a3_cov/mean(a3_cov)*prefac
		new_a3_cov = new_a3_cov/mean(new_a3_cov)*prefac

		if(a3_cov == new_a3_cov)
			@printf(STDERR,"a3_cov does not change!\n")
			exit(7)
		end

		tp_cov     = cal_cov(a3_cov,seq,para,genelen,extended_len,seq_score_mat)
		new_tp_cov = cal_cov(new_a3_cov,seq,para,genelen,extended_len,seq_score_mat)

		if(tp_cov == new_tp_cov)
			@printf(STDERR,"tp_cov does not change!\n")
			exit(8)
		end

		tp_cov = tp_cov/mean(tp_cov)
		new_tp_cov = new_tp_cov/mean(new_tp_cov)

		if(tp_cov == new_tp_cov)
			@printf(STDERR,"tp_cov does not change!\n")
			exit(9)
		end

		energy = -log(sum(sqrt(tp_cov.*tp_cov_real)))
		new_energy = -log(sum(sqrt(new_tp_cov.*tp_cov_real)))

		if new_energy == energy
			@printf(STDERR,"new_energy == energy!\n")
			exit(10)
		elseif new_energy < energy
			a3_cov = new_a3_cov
		elseif rand() < exp( sa_prefac*(energy-new_energy)/t)
			a3_cov = new_a3_cov
		end

		niter+=1
	end
end

tp_cov = tp_cov/mean(tp_cov)
a3_cov = a3_cov/mean(a3_cov)
